using System;
using System.Runtime.Serialization;
using System.Collections.Generic;
using Newtonsoft.Json;


namespace IMelody
{
    public class Melody
    {
        public string MelodyName;
        public float Speed;
        public float Ttactsize1;
        public float Ttactsize2;

        public List<MelodyNote> notes = new List<MelodyNote>();

        public struct MelodyNote {
            public int MidiNote;
            public float Divider;
            public int CaptureXpos;
            public int CaptureStaff;
            public int Tact;
            public float StartTime;
            public int  Direction;
            public int Dot;
        }

//        public Melody() {
//            Notes = new MelodyNote[1];
//        }

        public Melody Load(string tFile) {
            System.IO.StreamReader file2 = new System.IO.StreamReader(tFile);
            string json = file2.ReadToEnd();
            return( JsonConvert.DeserializeObject<Melody>(json));
        }

        public void Save(string tFile) {

            string js_out = JsonConvert.SerializeObject(this, Formatting.Indented);
            System.IO.StreamWriter file1 = new System.IO.StreamWriter(tFile);
            file1.Write(js_out);
            file1.Flush();
            file1.Close();
        }  // and of save

        public int GetMaxStaff() {
            int maxStaff = 0;
            for (int i = 0; i < notes.Count; i++) { 
                if( notes[i].CaptureStaff > maxStaff){
                    maxStaff = notes[i].CaptureStaff;
                };
            };
            return maxStaff;
        }

        public int GetMaxCaptureXPos(int staff)
        {
            int maxPos = 0;
            for (int i = 0; i < notes.Count; i++)
            {
                if ( notes[i].CaptureStaff == staff &&   notes[i].CaptureXpos > maxPos)
                {
                    maxPos = notes[i].CaptureXpos;
                };
            };
            return maxPos;
        }
        public void SetCapturePosX(int noteindex, int posX) {
            MelodyNote tmp = notes[noteindex];
            tmp.CaptureXpos = posX;
            notes[noteindex] = tmp;
        }

        public float GetMaxDivider(int staff, int posBegin, int posEnd) {
            float maxDivider = 0;
            for (int i = 0; i < notes.Count; i++)
            {
                if (notes[i].CaptureStaff == staff && notes[i].CaptureXpos > posBegin && notes[i].CaptureXpos < posEnd + 1  && notes[i].MidiNote < 1000 && notes[i].Divider > maxDivider)
                {
                    maxDivider = notes[i].Divider;
                };
            };
            return maxDivider;
        
        }

        public int GetTactLinePos(int staff,int startpos) {
            int minLinePos = GetMaxCaptureXPos(staff) ; 
            for (int i = 0; i < notes.Count; i++) {
                if (notes[i].CaptureStaff == staff && notes[i].CaptureXpos > startpos &&notes[i].MidiNote == 1000) {
                    if (notes[i].CaptureXpos < minLinePos) {
                        minLinePos = notes[i].CaptureXpos;
                    };
                };
            };
            return minLinePos;
        }
        
        public int GetNoteMinPosInTact(int staff, int posBegin, int posEnd){
            int minPos = posEnd;
            for (int i = 0; i < notes.Count; i++)
            {
                if (notes[i].CaptureStaff == staff && notes[i].CaptureXpos >= posBegin && notes[i].CaptureXpos <= posEnd && notes[i].MidiNote < 1000)
                {
                    if (notes[i].CaptureXpos < minPos)
                    {
                        minPos = notes[i].CaptureXpos;
                    };
                };
            };
            return minPos;
        }

        public int GetNoteMaxPosInTact(int staff, int posBegin, int posEnd)
        {
            int minPos = 0;
            for (int i = 0; i < notes.Count; i++)
            {
                if (notes[i].CaptureStaff == staff && notes[i].CaptureXpos > posBegin && notes[i].CaptureXpos < posEnd && notes[i].MidiNote < 1000)
                {
                    if (notes[i].CaptureXpos > minPos)
                    {
                        minPos = notes[i].CaptureXpos;
                    };
                };
            };
            return minPos;
        }

        public void AddNote(int tMidiNote, float tDivider, int tCaptureXpos, int tTact, float tStartTime, int tDirection = 0, int tDot = 0, int tCaptureStaff = 0)
        {

            if (tMidiNote < 1) {
                return;
            };

            MelodyNote tmpNote = new MelodyNote();
            tmpNote.MidiNote = tMidiNote;
            tmpNote.Divider = tDivider;
            tmpNote.CaptureXpos = tCaptureXpos;
            tmpNote.Tact = tTact;
            tmpNote.StartTime = tStartTime;
            tmpNote.Direction = tDirection;
            tmpNote.Dot = tDot;
            tmpNote.CaptureStaff = tCaptureStaff;
           
            notes.Add(tmpNote);
        }

        public int FindCaptureNote(int xPos, int midinote, int staff){
            int retIndex = -1;
            if (notes == null) {
                return retIndex;
            };

            for (int i = 0; i < notes.Count; i++ )
            {
                if(notes[i].CaptureStaff == staff && Math.Abs(notes[i].CaptureXpos - xPos) < 15 && notes[i].MidiNote == midinote){
                    return i;
                };
            };

            return retIndex;
        }

        public int FindCaptureStepline(int xPos, int staff)
        {
            int retIndex = -1;
            if (notes == null)
            {
                return retIndex;
            };

            for (int i = 0; i < notes.Count; i++)
            {
                if (notes[i].CaptureStaff == staff && Math.Abs(notes[i].CaptureXpos - xPos) < 15 && (notes[i].MidiNote == 1001 || notes[i].MidiNote == 1002))
                {
                    return i;
                };
            };

            return retIndex;
        }

        public int GetTactCount() { 
          int tactcnt = 0;
            for( int i = 0; i < notes.Count; i++){
                if(notes[i].Tact > tactcnt){
                   tactcnt = notes[i].Tact; 
                };
            };
            return tactcnt;
        }

        public void Sort() {
            int flag = 1;
            while(flag == 1){
                flag = 0;
                for (int i = 0; i < notes.Count - 1; i++) {
                    if (notes[i].CaptureStaff * 100000 + notes[i].CaptureXpos > notes[i+1].CaptureStaff * 100000 + notes[i+1].CaptureXpos)
                    {
                        MelodyNote tmp = notes[i];
                        
                        notes[i] = notes[i + 1];
                        notes[i + 1] = tmp;
                        flag = 1;
                    }
              }
            }
        }

        public void PostProcess(int tactsize1, float tactsize2)
        {


            
            int tact = 1;
            Ttactsize1 = tactsize1;
            Ttactsize2 = tactsize2;

            // remove tact lines
            for (int i = notes.Count - 1; i >= 0; i--) {
                if (notes[i].MidiNote == 1001 || notes[i].MidiNote == 1002) {
                    notes.RemoveAt(i);
                };
            };

            Sort();

            for (int staff = 1; staff <= GetMaxStaff(); staff++) {
                int maxXpos = GetMaxCaptureXPos(staff);
                for (int tactLinePos = 0; tactLinePos < maxXpos; tactLinePos = GetTactLinePos(staff, tactLinePos)) {
                    CreateTactGrid(staff, tactLinePos, GetTactLinePos(staff, tactLinePos), tact, tactsize1, tactsize2);
                    tact++;
                };
            };

        } // end of postprocess

        void CreateTactGrid(int staff, int posBegin, int posEnd, int tact, int tactsize1, float tactsize2)
        { 
            int noteMinPos = GetNoteMinPosInTact(staff, posBegin, posEnd) ; 
            int noteMaxPos = GetNoteMaxPosInTact(staff, posBegin, posEnd+1) ; // if ==0 then EMPTY TACT
            if (noteMaxPos - noteMinPos < 100)
            {
                noteMaxPos = noteMinPos + 100;
            };
            float maxdiv = GetMaxDivider(staff, posBegin, posEnd); // if ==0 then EMPTY TACT

            //int tactsize = 3;
            //float tactsize2 = 4;

            if (maxdiv < Ttactsize2) { maxdiv = Ttactsize2; }; // for melody 4/4

            //float XOffset = ((float)(noteMaxPos - noteMinPos) / ((int)maxdiv-1) ) /2; // all notes around axis in this range
            float tactstep = tactsize2 /maxdiv;
            float numoflines = Ttactsize1 / tactstep;

            float XOffset = (float)(noteMaxPos - noteMinPos) / (numoflines-1); // all notes around axis in this range

            
            float start = 1;
            for (float z = 0; z < numoflines ; z++)
            {
               MelodyNote tmp = new MelodyNote();
                tmp.CaptureStaff = staff;
                tmp.CaptureXpos = (int)(z * XOffset) + noteMinPos;
                tmp.StartTime = start;
                tmp.MidiNote = 1001;
                tmp.Tact = tact;

                if (tmp.CaptureXpos < 0)
                {
                    int breacpoint = 1;
                };


                   if( start != (int)start ){
                       tmp.MidiNote = 1002;
                   };
                notes.Add(tmp);
                start = start + tactstep;
            };
        }

        public void FinishPostProcess() {
            Sort();
            for (int i = 0; i < notes.Count; i++) {
                if (notes[i].MidiNote < 1000) {
                    MelodyNote tmp = notes[i];

                    int TactLineIdx = GetTactLineIdx(tmp.CaptureStaff, tmp.CaptureXpos);

                    tmp.Tact = notes[TactLineIdx].Tact;
                    tmp.StartTime = notes[TactLineIdx].StartTime;
                    
                    notes[i] = tmp;

                };
            };
        }


        int GetTactLineIdx(int staff, int posX) {
            int idx = 0;
            int nearpos = 1000000;
            for (int i = 0; i < notes.Count; i++) {
                if (notes[i].CaptureStaff == staff && ( notes[i].MidiNote == 1001 || notes[i].MidiNote == 1002)  ) { 
                    int offs =  Math.Abs( notes[i].CaptureXpos - posX );
                    if (offs < nearpos) {
                        nearpos = offs;
                        idx = i;
                    };
                }
            };

            return idx;
        }

    } // end of class
}// end of namespace
